"""
Methods about time.
"""
import pprint
import time
import warnings
from collections import OrderedDict

from lumo.proc.date import strftime


def format_second(sec: int) -> str:
    """convert seconds from int to string"""
    sec, ms = divmod(sec, 1)
    if sec > 60:
        min, sec = divmod(sec, 60)
        if min > 60:
            hour, min = divmod(min, 60)
            fmt = "{}h{}m{}s".format(hour, min, int(sec))
        else:
            fmt = "{}m{}s".format(min, int(sec))
    else:
        fmt = "{}s".format(int(sec))
    return fmt


class TimeIt:
    """
    A class for timing the time cost in each part.

    A global object is contained in lumo:

    ```python
    from lumo.utils import timeit
    timeit.start()
    timeit.mart("load")
    timeit.end("load")
    print(timeit.meter())
    ```

    """

    def __init__(self):
        self.last_update = None
        self.ends = False
        self.times = OrderedDict()

    def offset(self):
        now = time.time()

        if self.last_update is None:
            offset = 0
        else:
            offset = now - self.last_update

        self.last_update = now
        return offset, now

    def clear(self):
        self.last_update = None
        self.ends = False
        self.times.clear()

    def start(self):
        self.clear()
        self.mark("start", True)

    def mark(self, key, add_now=False):
        if self.ends:
            warnings.warn("called end method, please use start to restart timeit")
            return
        key = str(key)
        offset, now = self.offset()

        self.times.setdefault("use", 0)
        self.times["use"] += offset

        if add_now:
            self.times[key] = strftime("%H:%M:%S")
        else:
            self.times.setdefault(key, 0)
            self.times[key] += offset

    def end(self):
        self.mark("end", True)
        self.ends = True

    def meter(self, ratio=True):
        from lumo import Meter

        meter = Meter()
        for key, offset in self.times.items():
            if ratio:
                if isinstance(offset, str):
                    continue
                meter[key] = offset / self.times['use']
            else:
                meter[key] = offset

        return meter

    def __str__(self):
        return pprint.pformat(self.times)

    def __getitem__(self, item):
        return self.times[item]

    def __getattr__(self, item):
        return self.times[item]


timeit = TimeIt()



from typing import Tuple, Union, List

from torch import nn
from torch.nn.modules.batchnorm import _BatchNorm


def walk_module(module: Union[Tuple[nn.Module, ...], nn.Module]) -> Tuple[nn.Module, str, nn.Parameter]:
    if isinstance(module, (tuple, list)):
        for item in module:
            for ssubmodule, subname, subparam in walk_module(item):
                yield ssubmodule, subname, subparam
    else:
        for name, submodule in module.named_children():
            for ssubmodule, subname, subparam in walk_module(submodule):
                yield ssubmodule, subname, subparam

        for pname, param in module.named_parameters(recurse=False):
            yield module, pname, param


class ParamGrouper:
    """
    Most of pytorch Optimizers have the hyperparam `weight decay`, but if you use `model.parameters()` as the
     Optimizer's parameter, it will decay all parameters including bias,

    《Bag of Tricks for Image Classification with Convolutional Neural Networks》 suggests that only
    decay weight of linear or convolution layer.

    The best practice to do this in pytorch is to build param_groups for the parameter of Optimizer, and this class
    can do this easily and elegant.

    Examples:
    >>> grouper = ParamGrouper(model)
    ... noptim = params.optim.args.copy()
    ... noptim['weight_decay'] = 0
    ... param_groups = [
    ...     grouper.create_param_group(grouper.kernel_params(with_norm=False), **params.optim.args),
    ...     grouper.create_param_group(grouper.bias_params(with_norm=False), **noptim),
    ...     grouper.create_param_group(grouper.norm_params(), **noptim),
    ... ]
    """

    def __init__(self, *module: nn.Module):
        self.module = module

    def params(self, with_norm=True):
        params = []
        for module, name, param in walk_module(self.module):
            if with_norm or not isinstance(module, (_BatchNorm, nn.LayerNorm)):
                params.append(param)
        return params

    def kernel_params(self, with_norm=False):
        params = []
        for module, name, param in walk_module(self.module):
            if 'weight' in name and (with_norm or not isinstance(module, (_BatchNorm, nn.LayerNorm))):
                params.append(param)
        return params

    def bias_params(self, with_norm=False):
        params = []
        for module, name, param in walk_module(self.module):
            if 'bias' in name and (with_norm or not isinstance(module, (_BatchNorm, nn.LayerNorm))):
                params.append(param)
        return params

    def batchnorm_params(self):
        return [param for module, name, param in walk_module(self.module) if isinstance(module, _BatchNorm)]

    def layernorm_params(self):
        return [param for module, name, param in walk_module(self.module) if isinstance(module, nn.LayerNorm)]

    def norm_params(self):
        return [param for module, name, param in walk_module(self.module) if
                isinstance(module, (_BatchNorm, nn.LayerNorm))]

    @staticmethod
    def create_param_group(params, **kwargs):
        kwargs['params'] = params
        return kwargs



"""

"""

from lumo.base_classes.trickitems import NoneItem, AvgItem


def test_none_item():
    none = NoneItem()
    assert none > 0
    assert none >= 0
    assert none < 0
    assert none <= 0
    assert none != 0
    assert none == None
    assert none is not None
    assert none / 2 == 0.5
    assert none * 2 == 2
    assert none + 1 == 1
    assert none - 1 == -1


def test_avg_item():
    avg = AvgItem()
    avg.update(4)
    avg.update(3)
    avg.update(2)
    assert avg._item == 2
    assert avg.avg == 3



"""

"""

import torch


class BatchEmbedding():
    """用于将多个 batch 的输入一同添加到 tensorboard 中"""
    def __init__(self, writer, global_step=None, tag="default"):
        self.writer = writer
        self.global_step = global_step
        self.tag = tag
        self.mats = []
        self.metadatas = []
        self.label_imgs = []

    def add_embedding(self, mat, metadata=None, label_img=None):
        self.mats.append(mat.detach().cpu())
        self.metadatas.append(metadata.detach().cpu())
        self.label_imgs.append(label_img.detach().cpu())

    def flush(self, max_len=None):
        self.writer.add_embedding(torch.cat(self.mats[:max_len]),
                                  torch.cat(self.metadatas[:max_len]),
                                  self.label_imgs,
                                  tag=self.tag,
                                  global_step=self.global_step)




from typing import Union, NoReturn, Dict, overload
from lumo.base_classes import TrainerStage
from torch.utils.data import DataLoader, Dataset
from .mixin import DataModuleMix
from .builder import DatasetWrap
from .params import ParamsType


class DataModule(DataModuleMix):
    train = TrainerStage.train
    test = TrainerStage.test
    val = TrainerStage.val
    debug = TrainerStage.debug

    def __init__(self, train: DataLoader = None, val: DataLoader = None, test: DataLoader = None):
        self._dataloader = {}
        self.regist_dataloader(train=train, val=val, test=test)

    def __getitem__(self, item):
        return self._dataloader.get(item, None)

    @property
    def train_dataloader(self) -> Union[NoReturn, DataLoader]:
        return self['train']

    @property
    def test_dataloader(self) -> Union[NoReturn, DataLoader]:
        return self['test']

    @property
    def val_dataloader(self) -> Union[NoReturn, DataLoader]:
        return self['val']

    @overload
    def regist_dataloader(self, train: DataLoader = None, val: DataLoader = None, test: DataLoader = None,
                          **others: Dict[str, DataLoader]):
        ...

    def regist_dataloader(self, **kwargs: dict):
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        self._dataloader.update(kwargs)

    def regist_dataloader_with_stage(self, stage: TrainerStage, loader: DataLoader):
        kwargs = {stage.name: loader}
        self.regist_dataloader(**kwargs)

    def idataloader(self, params: ParamsType, stage: TrainerStage, repeat: bool = False):
        pass

    @classmethod
    def wrap_dataloader(cls, dataset: Dataset) -> DataLoader:
        self = cls()
        self._dataset = dataset
        dataset = DatasetWrap(dataset)
        return dataset.DataLoader()



"""
"""
from torchvision.models.resnet import resnet34
from torch.nn import functional as F
from torch import nn
from lumo import Trainer, Params, DatasetBuilder, DataModule, callbacks, Meter
from lumo.base_classes import TrainerStage
from lumo.proc import path


class DemoPM(Params):

    def __init__(self):
        super().__init__()
        self.epoch = 20
        self.batch_size = 128
        self.num_workers = 8
        self.pin_memory = True

        self.pretrain = True
        self.pretrain_path = None
        self.dataset = self.choice('cifar10', 'cifar100')
        self.optim = self.OPTIM.create_optim('SGD', weight_decay=0.001, lr=0.01, momentum=0.9)
        self.dataset_root = path.dataset_cache_dir('demo')
        self.download_dataset = True
        self.lr_sche = None
        self.n_classes = 10

    def iparams(self):
        super().iparams()
        if self.dataset == 'cifar100':
            self.n_classes = 100

        self.lr_sche = self.SCHE.Cos(start=0, end=self.epoch,
                                     left=self.optim.args.lr,
                                     right=1e-6)


ParamsType = DemoPM


class CVDM(DataModule):

    def idataloader(self, params: ParamsType, stage: TrainerStage, repeat: bool = False):
        from torchvision.datasets import cifar
        if params.dataset == 'cifar10':
            dataset = cifar.CIFAR10(params.dataset_root, train=stage.is_train, download=params.download_dataset)
        else:
            dataset = cifar.CIFAR100(params.dataset_root, train=stage.is_train, download=params.download_dataset)

        ds = (
            DatasetBuilder()
                .add_input('xs', dataset.data)
                .add_input('ys', dataset.targets)
                .add_output('xs', 'xs')
                .add_output('ys', 'ys')
        )
        loader = ds.DataLoader(batch_size=params.batch_size,
                               num_workers=params.num_workers,
                               pin_memory=params.pin_memory)

        self.regist_dataloader_with_stage(stage, loader)


# Trainer


class ResnetTrainer(callbacks.TrainCallback,
                    Trainer):
    def icallbacks(self, params: ParamsType):
        callbacks.LoggerCallback(step_frequence=1).hook(self)
        callbacks.LRSchedule().hook(self)  # auto get params.lr_sche to apply lr rate
        callbacks.AutoLoadModel().hook(self)

        if isinstance(self, callbacks.TrainCallback):
            self.hook(self)

    def imodels(self, params: ParamsType):
        super().imodels(params)

        model = resnet34()
        model.fc = nn.Linear(1024, params.n_classes)
        self.model = model
        self.optim = params.optim.build(self.model.parameters())
        self.to_device()

    def train_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        meter = Meter()
        xs, ys = batch['xs'], batch['ys']
        logits = self.model(xs)
        loss = F.cross_entropy(logits, ys)

        meter.Lce = loss
        meter.Acc = (logits.argmax(dim=-1) == ys).float().mean()

        self.optim.zero_grad()
        self.accelerator.backward(loss)
        self.optim.step()
        return meter


def main():
    pm = DemoPM()
    pm.from_args()

    dm = CVDM()

    tr = ResnetTrainer(pm)
    tr.train(dm)
    tr.save_model()


if __name__ == '__main__':
    main()



from fire import Fire


class Interface:
    def init(self):
        pass

    def summary(self):
        pass

    def list(self):
        pass

    def latest(self):
        pass

    def info(self, test_name):
        pass

    def clear(self, cache=True):
        pass


if __name__ == '__main__':
    Fire(Interface)



"""
一个训练 f(x) = x+1 的线性函数的例子
"""
from lumo import Trainer, DatasetBuilder, Params, callbacks
from lumo.kit.params import ParamsType

import torch
from torch import nn

from lumo import Meter


class PlusOneTrainer(Trainer):

    def imodels(self, params: ParamsType):
        self.model = nn.Linear(1, 1)
        self.optim = params.optim.build(self.model.parameters())
        self.to_device()

    def icallbacks(self, params: ParamsType):
        callbacks.LoggerCallback().hook(self)

    def train_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        meter = Meter()
        xs, ys = batch
        logits = self.model(xs.float())
        loss = torch.mean((ys - logits) ** 2)
        self.optim.zero_grad()
        loss.backward()
        self.optim.step()
        meter.Lce = loss
        return meter


params = Params()
params.epoch = 30
params.batch_size = 10
params.optim = params.OPTIM.create_optim('SGD', lr=0.000001, weight_decay=4e-3)

trainer = PlusOneTrainer(params)

builder = (
    DatasetBuilder().add_input('xs', range(-500, 500)).add_input('ys', range(-500, 500))
        .add_output('xs', 'xs').add_output('ys', 'ys')
        .add_output_transform('xs', lambda x: torch.tensor([x]))
        .add_output_transform('ys', lambda x: torch.tensor([x + 1]))
        .random_sampler().chain()
)

builder = (
    DatasetBuilder(
        xs=range(-500, 500),
        ys=range(-500, 500),
        auto_output=True,
        xs_opt_transform=lambda x: torch.tensor([x]),
        ys_opt_transform=lambda x: torch.tensor([x + 1])
    ).random_sampler().chain()
)  # another simpler way

trainer.train(builder.DataLoader(batch_size=params.batch_size, num_workers=4))

print(list(trainer.model.parameters()))



import json
import sqlite3
import time
import random
import sys
import os

from lumo.base_classes import attr
from lumo.utils.hash import string_hash
from lumo.proc.path import cache_dir

class _NoneResult():
    def fetchone(self):
        return None


class MPStruct:
    TABLE_NAME = None
    TABLE_SQL = None
    INDEX_SQL = None

    def __init__(self, session_id=None, root=None, retry=7):
        if session_id is None:
            session_id = string_hash(sys.argv[0])

        if root is None:
            root = os.path.join(cache_dir(), '.mpqueue')
            os.makedirs(root, exist_ok=True)
        self.file = os.path.join(root, f"{self.__class__.__name__}_{session_id}.sqlite")
        self.root = root
        self._retry = retry
        self.session_id = session_id
        self._connect = None
        self._initialized = False
        self._tables = None
        self._pop_queue = None

    @property
    def connect(self):
        if self._connect is None:
            self._connect = sqlite3.connect(self.file)
        return self._connect

    def reconnect(self):
        self._connect.close()
        self._connect = sqlite3.connect(self.file)

    @property
    def cursor(self):
        if not self._initialized:
            self.init_table()
            self._initialized = True
        return self.connect.cursor()

    def init_table(self):
        raise NotImplemented()

    def has_table(self, key):
        if self._tables is None:
            conn = self.connect
            c = conn.cursor()
            self._tables = {i[0] for i in
                            c.execute("""SELECT name FROM sqlite_master WHERE type='table';""").fetchall()}
        return key in self._tables

    def encode_value(self, value):
        res = attr()
        res.value = value
        return json.dumps(res.jsonify())

    def decode_value(self, res):
        res = json.loads(res)
        return attr.from_dict(res).value

    def execute(self, sql, mode='r'):
        for i in range(self._retry):
            try:
                res = self.cursor.execute(sql)
                if mode != 'r':
                    self.connect.commit()
                return res
            except sqlite3.OperationalError as e:
                from lumo.kit.logger import get_global_logger
                self.connect.rollback()
                get_global_logger().warn(f'[mpqueue] retry {i:02d}/{self._retry}...', sql, e)
                time.sleep(((i + 1) ** 2) * random.random())
                continue
        return _NoneResult()

    def executescript(self, sql, mode='r'):
        for i in range(self._retry):
            try:
                res = self.cursor.executescript(sql)
                if mode != 'r':
                    self.connect.commit()
                return res
            except sqlite3.OperationalError as e:
                from lumo.kit.logger import get_global_logger
                get_global_logger().warn(f'[mpqueue] retry {i:02d}/{self._retry}...', sql, e)
                # self.reconnect()
                time.sleep(random.random() * i + 0.5)
                continue
        return _NoneResult()



from setuptools import setup, find_packages
from lumo import __version__

# from rename_templete import copy_templete

# copy_templete()

"""
python3 setup.py sdist bdist_wheel; 
python3 setup.py sdist bdist_wheel; sudo pip install dist/$(python3 install.py);
python3 setup.py sdist bdist_wheel; pip install dist/$(python3 install.py) --user
sudo pip install dist/$(python3 install.py);
pip install dist/$(python3 install.py) --user
"""

setup(
    name='lumo',
    version=__version__,
    description='torch kit for programing your dl experiments code elegant.',
    url='https://github.com/sailist/lumo',
    author='sailist',
    author_email='sailist@outlook.com',
    license='Apache License 2.0',
    include_package_data=True,
    install_requires=[
        'fire', 'psutil', 'joblib'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='lumo',
    packages=['lumo'],
    entry_points={
        'console_scripts': [
            'lumo = lumo.cli.cli:main'
        ]
    },
)



"""

"""
import sys
sys.path.insert(0,"../")
from lumo import __version__
print(__version__)


import time
from lumo import Logger,Meter,Params
logger = Logger()

meter = Meter()
meter.a = 3.13524635465
meter.b = "long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;long text;"

for i in range(10):
    logger.inline("inline examples",meter)
    time.sleep(0.2)
logger.info(1,2,3,{4:4})
logger.info(meter)

for i in range(10):
    logger.raw(i,inline=True,append=True)
    time.sleep(0.2)

logger.add_log_dir("./")


"""
used for logging
"""

import os
import sys
from collections import namedtuple
from datetime import datetime
from typing import Any, Callable

from lumo.proc.date import strftime
from ..utils.screen import ScreenStr

import logging

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)

loginfo = namedtuple('loginfo', ['string', 'prefix_len'])

logger = None

VVV_DEBUG = -10
VV_DEBUG = 0
V_DEBUG = 10
V_INFO = 20
V_WARN = 30
V_ERROR = 40
V_FATAL = 50


def get_global_logger():
    global logger
    if logger is None:
        logger = Logger()
    return logger


def set_global_logger(logger_):
    global logger
    logger = logger_
    return logger


class Logger:
    VERBOSE = 20
    VVV_DEBUG = VVV_DEBUG
    VV_DEBUG = VV_DEBUG
    V_DEBUG = V_DEBUG
    V_INFO = V_INFO
    V_WARN = V_WARN
    V_ERROR = V_ERROR
    V_FATAL = V_FATAL
    _instance = None

    def __new__(cls, *args, **kwargs) -> Any:
        if Logger._instance is not None:
            return Logger._instance
        return super().__new__(cls)

    def __init__(self, adddate=True, datefmt: str = '%y-%m-%d %H:%M:%S', sep: str = " | ", use_stdout: bool = True):
        if Logger._instance is not None:
            return

        self.adddate = adddate
        self.datefmt = datefmt
        self.out_channel = []
        self.pipe_key = set()
        self.sep = sep
        self.return_str = ""
        self.listener = []
        self.use_stdout = use_stdout
        Logger._instance = self

    def _format(self, *values, inline=False, fix=0, raw=False, append=False, level=V_INFO):
        """"""
        if level < Logger.VERBOSE:
            return None, None

        if self.adddate and not raw:
            cur_date = datetime.now().strftime(self.datefmt)
        else:
            cur_date = ""

        values = ["{}".format(str(i)) for i in values]
        values = [i for i in values if len(i.strip()) != 0]

        if len(cur_date) == 0:
            space = [*["{}".format(str(i)) for i in values]]
        else:
            space = [cur_date, *values]

        if fix >= 0:
            left, right = self.sep.join(space[:fix + 1]), self.sep.join(space[fix + 1:])
            fix = len(left) + len(self.sep)
            logstr = left
            if len(right) > 0:
                logstr = self.sep.join((logstr, right))

            if inline:
                if append:
                    return "{}".format(logstr), fix
                else:
                    return "\r{}".format(logstr), fix
            else:
                return "{}\n".format(logstr), fix

        space = self.sep.join(space)

        if inline:
            return loginfo("\r{}".format(space), 0)
        else:
            return loginfo("{}\n".format(space), 0)

    def _handle(self, logstr, end="", level=10, **kwargs):
        """handle log stinrg"""
        for listener in self.listener:
            listener(logstr, end, level)

        if level < Logger.VERBOSE:
            return
        file = sys.stdout
        if level > Logger.V_INFO:
            file = sys.stderr

        if logstr.startswith("\r"):
            fix = kwargs.get("fix", 0)
            self.return_str = logstr
            self.print(ScreenStr(logstr, leftoffset=fix), end=end, file=file)
        else:
            if len(self.return_str) != 0:
                self.print(self.return_str, end="\n", file=file)
            self.print(logstr, end="", file=file)

            for i in self.out_channel:
                with open(i, "a", encoding="utf-8") as w:
                    if len(self.return_str) != 0:
                        w.write("{}\n".format(self.return_str.strip()))
                    w.write(logstr)

            self.return_str = ""

    def inline(self, *values, fix=0, append=False):
        """Log a message with severity 'INFO' inline"""
        logstr, fix = self._format(*values, inline=True, fix=fix, append=append)
        if fix is None:
            return
        self._handle(logstr, level=V_INFO, fix=fix)

    def info(self, *values):
        """Log a message with severity 'INFO'"""
        logstr, fix = self._format(*values, inline=False)
        if fix is None:
            return
        self._handle(logstr, level=V_INFO, fix=fix)

    def raw(self, *values, inline=False, fix=0, level=20, append=False):
        """Log a message with severity 'INFO' withou datetime prefix"""
        logstr, fix = self._format(*values, inline=inline, fix=fix, raw=True, append=append)
        if fix is None:
            return
        self._handle(logstr, level=level)

    def debug(self, *values):
        """Log a message with severity 'DEBUG'"""
        logstr, fix = self._format("DEBUG", *values, inline=False)
        if fix is None:
            return
        self._handle(logstr, level=V_DEBUG, fix=fix)

    def ddebug(self, *values):
        """Log a message with severity 'DEBUG'"""
        logstr, fix = self._format("DDEBUG", *values, inline=False)
        if fix is None:
            return
        self._handle(logstr, level=VV_DEBUG, fix=fix)

    def dddebug(self, *values):
        """Log a message with severity 'DEBUG'"""
        logstr, fix = self._format("DDDEBUG", *values, inline=False)
        if fix is None:
            return
        self._handle(logstr, level=VVV_DEBUG, fix=fix)

    def warn(self, *values):
        """Log a message with severity 'WARN'"""
        logstr, fix = self._format("WARN", *values, inline=False)
        if fix is None:
            return
        self._handle(logstr, level=V_WARN, fix=fix)

    def error(self, *values):
        """Log a message with severity 'ERROR'"""
        logstr, fix = self._format("ERROR", *values, inline=False)
        if fix is None:
            return
        self._handle(logstr, level=V_ERROR, fix=fix)

    def fatal(self, *values):
        """Log a message with severity 'FATAL'"""
        logstr, fix = self._format("FATAL", *values, inline=False)
        if fix is None:
            return
        self._handle(logstr, level=V_FATAL, fix=fix)

    def newline(self):
        """break line"""
        self._handle("", level=V_INFO)

    def print(self, *args, end='\n', file=sys.stdout):
        """built-in print function"""
        if self.use_stdout:
            print(*args, end=end, flush=True, file=file)

    def toggle_stdout(self, val: bool = None):
        """False will stop write on stdout"""
        if val is None:
            val = not self.use_stdout
        self.use_stdout = val

    def add_log_dir(self, dir):
        """add a file output pipeline"""
        if dir in self.pipe_key:
            self.info("Add pipe {}, but already exists".format(dir))
            return None

        os.makedirs(dir, exist_ok=True)

        i = 0
        cur_date = strftime(fmt="%y%m%d%H%M")
        fni = os.path.join(dir, f"l.{i}.{cur_date}.log")
        while os.path.exists(fni):
            i += 1
            fni = os.path.join(dir, "l.{}.{}.log".format(i, cur_date))

        self.print("add output channel on {}".format(fni))
        self.out_channel.append(fni)
        self.pipe_key.add(dir)
        return fni

    def add_log_listener(self, func: Callable[[str, str, int], Any]):
        """add a log event handler"""
        self.listener.append(func)

    def set_verbose(self, verbose=20):
        """set log verbose, default level is `INFO`"""
        Logger.VERBOSE = verbose
        logging.basicConfig(format='%(levelname)s:%(message)s', level=verbose)



"""
Methods about git.
"""
import git
import os
from functools import lru_cache
from lumo.base_classes import attr

try:
    from git import Repo, Commit
except ImportError:
    def Repo(*_, **__):
        return None


    def Commit(*_, **__):
        return None

from lumo.proc.explore import git_enable
from lumo.proc.path import git_dir, cache_dir
from lumo.utils.paths import checkpath


class GitKey:
    LUMO_BRANCH = 'lumo_experiments'


class GitCMDResult(attr):
    def __init__(self, cmd, result=0, msg=''):
        self.cmd = cmd
        self.result = result
        self.msg = msg

    @classmethod
    def failed(cls, cmd, msg):
        return cls(cmd, 1, msg)


class branch:
    """
    用于配合上下文管理切换 git branch

    with branch(repo, branch):
        repo.index.commit('...')
    """

    def __init__(self, repo, branch: str):
        self.repo = repo
        self.old_branch = self.repo.head.reference
        self.branch = branch

    def __enter__(self):
        if self.branch not in self.repo.heads:
            head = self.repo.create_head(self.branch)
        else:
            head = self.repo.heads[self.branch]

        self.repo.head.reference = head
        # self.repo.head.reset(index=True, working_tree=True)
        return head

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.repo.head.reference = self.old_branch


def check_have_commit(repo):
    if len(repo.heads) == 0:
        repo.git.add('.')
        repo.index.commit('initial commit')


@lru_cache()
def load_repo(dir='./'):
    """
    Try to load git repository object of a directory.
    Args:
        dir: str, a directory path, default is the current working dir.
        if dir is a repository dir, then a git.Repo object will be retured.
        if not, some you can type a path to init it, or type '!' to cancel init it.

    Returns:
        git.Repo object or None if dir not have git repository and cancel to init it.
    """
    path = git_dir(dir)
    repo = Repo(path)
    check_have_commit(repo)
    return repo


class GitWrap:
    def add(self, repo=None):
        raise NotImplementedError()

    def commit(self, repo=None, key=None, branch_name='lumo_experiments', info: str = None):
        raise NotImplementedError()

    def reset(self):
        raise NotImplementedError()

    def archive(self):
        raise NotImplementedError()


class GitDisabledWrap(GitWrap):
    def add(self, repo=None):
        return GitCMDResult('add', 1, 'git cannot be used.')

    def commit(self, repo=None, key=None, branch_name='lumo_experiments', info: str = None):
        return GitCMDResult('commit', 1, 'git cannot be used.')

    def reset(self):
        return GitCMDResult('reset', 1, 'git cannot be used.')

    def archive(self):
        return GitCMDResult('archive', 1, 'git cannot be used.')


class GitEnabledWrap(GitWrap):

    def add(self, repo=None):
        if repo is None:
            repo = load_repo()
        return repo.git.add(all=True)

    def commit(self, repo=None, key=None, branch_name=GitKey.LUMO_BRANCH, info: str = None):
        """
        ```
            TODO behavior need to be verified.
            cd <repo working dir>
            git reset <branch_name>
            git add .
            git commit -m "<info>"
            git reset <original branch>
        ```
        Args:
            repo:
            key:
                to avoid duplicate commit, a key value can be passed,
                commit operation will be perform if `key` hasn't appeared before.
                default value is None, means you can commit as you want without limitation
            branch_name:
                commit on which branch, nonexistent branch will be created.
            info:
                commit info, a string
        Returns:
            git.Commit object, see gitpython for details.
        """
        try:
            if repo is None:
                repo = load_repo()

            if key is not None and key in _commits_map:
                return _commits_map[key]

            with branch(repo, branch_name):
                repo.git.add(all=True)
                commit_info = '[[EMPTY]]'
                if info is not None:
                    commit_info = info
                commit_ = repo.index.commit(commit_info)
            if key is not None:
                _commits_map[key] = commit_
        except git.GitCommandError:
            commit_ = None
        return commit_

    def reset(self, repo=None, commit_hex=None, commit=None):
        """
        TODO
        将工作目录中的文件恢复到某个commit
        恢复快照的 git 流程:
            git branch experiment
            git add . & git commit -m ... // 保证文件最新，防止冲突报错，这一步由 Experiment() 代为完成
            git checkout <commit-id> // 恢复文件到 <commit-id>
            git checkout -b reset // 将当前状态附到新的临时分支 reset 上
            git branch experiment // 切换回 experiment 分支
            git add . & git commit -m ... // 将当前状态重新提交到最新
                // 此时experiment 中最新的commit 为恢复的<commit-id>
            git branch -D reset  // 删除临时分支
            git branch master // 最终回到原来分支，保证除文件变动外git状态完好
        Returns:
            An Experiment represents this reset operation
        """
        commit = commit
        if repo is None:
            repo = load_repo()
        old_path = os.getcwd()
        os.chdir(commit.tree.abspath)
        # exp = Experiment('Reset')

        from thexp.utils.repository import branch
        with branch(commit.repo, GitKey.LUMO_BRANCH) as new_branch:
            repo.git.checkout(commit.hexsha)
            repo.git.checkout('-b', 'reset')
            repo.head.reference = new_branch
            repo.git.add('.')
            ncommit = repo.index.commit("Reset from {}".format(commit.hexsha))
            repo.git.branch('-d', 'reset')
        # exp.add_plugin('reset', {
        #     'test_name': self.name,  # 从哪个状态恢复
        #     'from': exp.commit.hexsha,  # reset 运行时的快照
        #     'where': commit.hexsha,  # 恢复到哪一次 commit，是恢复前的保存的状态
        #     'to': ncommit.hexsha,  # 对恢复后的状态再次进行提交，此时 from 和 to 两次提交状态应该完全相同
        # })
        # exp.end()

        os.chdir(old_path)
        return None

    def archive(self, repo=None, commit_hex=None, commit=None, tgt=None):
        """
        TODO
        将某次 test 对应 commit 的文件打包，相关命令为
            git archive -o <filename> <commit-hash>
        Returns:
            An Experiment represents this archive operation
        """
        if repo is None:
            repo = load_repo()

        old_path = os.getcwd()
        os.chdir(commit.tree.abspath)
        # exp = Experiment('Archive')

        revert_path = checkpath(cache_dir(), 'archives', commit)
        revert_fn = os.path.join(revert_path, "code.zip")

        # TODO 在code.zip目录下添加相关说明
        # exp.add_plugin('archive', {'file': revert_fn,
        #                            'test_name': self.name})
        with open(revert_fn, 'wb') as w:
            repo.archive(w, commit)

        # exp.end()
        os.chdir(old_path)
        return None


if git_enable():
    wrap = GitEnabledWrap()
else:
    wrap = GitDisabledWrap()

add = wrap.add
commit = wrap.commit
reset = wrap.reset
archive = wrap.archive

_commits_map = {}



from typing import List, Dict, Tuple, Union, overload, Optional, Sequence
from decorator import decorator
from torch import nn
import torch
from collections.abc import Sequence, Mapping

from torch.utils.data import DataLoader


def _to_device(item: Union[Sequence, Mapping, torch.Tensor, nn.Module],
               device_args_kargs: Tuple[Sequence, Mapping]):
    if isinstance(item, nn.Module):
        return item.to(*device_args_kargs[0], **device_args_kargs[1])

    if isinstance(item, Sequence):
        if isinstance(item, str):
            return item
        return [_to_device(ele, device_args_kargs) for ele in item]
    elif isinstance(item, Mapping):
        return {k: _to_device(ele, device_args_kargs) for k, ele in item.items()}
    elif isinstance(item, torch.Tensor):
        return item.to(*device_args_kargs[0], **device_args_kargs[1])
    else:
        raise TypeError(type(item))


def set_default_to_device_func(func):
    global to_device
    to_device = func


def get_to_device_func():
    return to_device


def to_device_enumrate(loader: DataLoader, device_args_kwargs: Tuple[Sequence, Dict]):
    to_device = get_to_device_func()
    for idx, batch in enumerate(loader):
        batch = to_device(batch, device_args_kwargs)
        yield idx, batch


@overload
def construct_device_args_kwargs(self, dtype: torch.dtype, non_blocking: bool = False, copy: bool = False):
    ...


@overload
def construct_device_args_kwargs(self, device: Optional[Union[torch.device, str]] = None,
                                 dtype: Optional[torch.dtype] = None,
                                 non_blocking: bool = False, copy: bool = False):
    ...


@overload
def construct_device_args_kwargs(self, other: torch.Tensor, non_blocking: bool = False, copy: bool = False):
    ...


def construct_device_args_kwargs(*args, **kwargs):
    return (args, kwargs)


try:
    from accelerate.data_loader import send_to_device
except:
    send_to_device = None


@decorator
def _wrap(func, item, device_arg):
    args, kwargs = device_arg
    device = kwargs.get('device', None)
    if device is None:
        device = args[0]
    return func(item, device)


if send_to_device is not None:
    to_device = _wrap(send_to_device)
else:
    to_device = _to_device



"""

"""
import random

import torch
from torch.nn import functional as F


class RandomHorizontalFlip():
    """Flip randomly the image.
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, x: torch.Tensor):
        """NCWH"""
        if random.random() < self.p:
            x = x.flip(dims=(-1,))
        return x.contiguous()


class RandomBatchCrop():
    def __init__(self, size, padding=4, fill=0, p=0.5):
        self.p = p
        self.size = size
        self.pad = Pad(padding, fill=fill)

    def __call__(self, x: torch.Tensor):
        """NCWH"""
        x = self.pad(x)

        h, w = x.shape[-2:]
        new_h = new_w = self.size

        top = random.randint(0, h - new_h)
        left = random.randint(0, w - new_w)

        x = x[:, :, top: top + new_h, left: left + new_w]
        return x


class Pad():
    def __init__(self, padding=4, fill=0):
        self.pad = padding
        self.fill = fill

    def __call__(self, x: torch.Tensor):
        x = F.pad(x, [self.pad] * 4)
        return x



from lumo.kit.mixin import BaseCallbackMix
from lumo.kit import callbacks
from lumo.kit.params import ParamsType


class BaseCallbackMixin(BaseCallbackMix):
    def icallbacks(self, params: ParamsType):
        params.step_frequence = params.default(1)
        callbacks.LoggerCallback(step_frequence=params.step_frequence).hook(self)
        callbacks.LRSchedule().hook(self)  # auto get params.lr_sche to apply lr rate
        callbacks.AutoLoadModel().hook(self)

        if isinstance(self, callbacks.TrainCallback):
            self.hook(self)



"""

"""
import textwrap
from pprint import pformat
from typing import List, Union, Any

import copy
import json
import os
import pprint as pp
import sys
import warnings
from collections import namedtuple
from collections.abc import Iterable
from datetime import timedelta
from itertools import chain
from typing import Any, overload, TypeVar, Optional, List, Union
from accelerate.kwargs_handlers import KwargsHandler
from accelerate.utils import RNGType
import fire
import torch

from lumo.base_classes.attr import attr
from lumo.base_classes.errors import BoundCheckError, NewParamWarning
from lumo.base_classes.params_vars import OptimBuilder, OptimMixin
from lumo.calculate import schedule
from lumo.utils import safe_io as io

arange_param = namedtuple('arange_param', ['default', 'left', 'right'], defaults=[None, float('-inf'), float('inf')])
choice_param = namedtuple('choice_param', ['default', 'choices'], defaults=[None, []])
default_param = namedtuple('default_param', ['default', 'warn'], defaults=[True])


class BaseParams():
    """
    Params make it easy to get/set/load/dump your config, if you use easy_dict before, you can see Params as a easy_dict ppplus.

    Notes:
        Variable name starts with `_` is not recommeded, because of the special inner implement, variable name starts with `_`
        will be ignored when serialize Params instnace, you will loss the value of these variables.

        If you really want to define some special variable to make it look different from others, you can capitalize it or ends with `_`.
    """

    def __init__(self):
        self._param_dict = attr()
        self._namespace = attr()
        self._repeat = -1
        self._constrain = {}
        self._lock = False

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        key = str(key)
        self.__setattr__(key, value)

    def __setattr__(self, name: str, value: Any) -> None:
        """
        1. check constrain
        2. check if is default and not exists
        """
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            if isinstance(value, (arange_param, choice_param)):
                self._constrain[name] = value
                value = value.default
            else:
                self._check(name, value)

            if isinstance(value, default_param):  # 设置默认值
                # only set default param when name not exists
                if name not in self._param_dict:
                    if value.warn:
                        warnings.warn(
                            "'{}' is a new param,please check your spelling. It's more recommended to define in advance.".format(
                                name))
                    value = value.default
                    self._param_dict[name] = value
            else:
                self._param_dict[name] = value

            res = self._namespace.setdefault(self.__class__, [])
            res.append(name)

    def __getattr__(self, item):
        if item not in self._param_dict and self._lock:
            raise AttributeError(item)
        return self._param_dict.__getattr__(item)

    def __delattr__(self, name: str) -> None:
        if name.startswith("_"):
            super().__delattr__(name)
        else:
            self._param_dict.pop(name)

    def __delitem__(self, key):
        key = str(key)
        self.__delattr__(key)

    def __contains__(self, item):
        return item in self._param_dict

    def __getstate__(self):
        return {
            '_param_dict': self._param_dict,
            '_repeat': self._repeat,
            '_lock': self._lock,
            '_bound': self._constrain,
            '_namespace': self._namespace,
        }

    def __setstate__(self, d):
        self._param_dict = d['_param_dict']
        self._repeat = d['_repeat']
        self._lock = d['_lock']
        self._constrain = d['_bound']
        self._namespace = d['_namespace']

    def __repr__(self):
        import textwrap

        dynamic_propertys = [(k, io.safe_getattr(self, k, None)) for k in self.__dir__() if
                             isinstance(getattr(self.__class__, k, None), property)]

        dynamic_propertys = [(k, v, f'{type(v).__name__}') for k, v in dynamic_propertys]

        def _arg_to_str(k, v):
            res = self._constrain.get(k, None)
            if res is not None:
                return f'{res}, {type(v).__name__}'
            return f'{type(v).__name__}'

        args = [(k, v) for k, v in
                chain(self._param_dict.items())]
        args = [(k, v, _arg_to_str(k, v)) for k, v in args]

        args_str = BaseParams.safe_param_repr(args)

        if len(dynamic_propertys) > 0:
            property_str = BaseParams.safe_param_repr(dynamic_propertys)
            return "{}.Space".format(
                self.__class__.__name__) + '(\n' + args_str + '\n    # @property\n' + property_str + '\n)'
        return "{}.Space".format(self.__class__.__name__) + '(\n' + args_str + '\n)'

    __str__ = __repr__

    def __eq__(self, other):
        """
        equal function only compare params, and will ignore other Params hidden variables, including `_repeat`, `_lock`, and
        `_bound`, so if param_1 equals to param_2, it means all key-value (need to be hashable) pair in these two Params
        instance is equal.
        """
        if isinstance(other, BaseParams):
            return self.hash() == other.hash()
        return False

    def _check(self, name, value):
        if isinstance(value, default_param):
            value = value.default
        if name not in self._constrain:
            return True
        bound = self._constrain[name]
        if isinstance(bound, arange_param) and not (bound.left <= value and value <= bound.right):
            raise BoundCheckError(f"value of param '{name}' should in range [{bound.left}, {bound.right}].")
        elif isinstance(bound, choice_param) and value not in bound.choices:
            raise BoundCheckError(f"value of param '{name}' should in values {bound.choices}.")

    @staticmethod
    def _safe_repr(values: Any) -> str:
        return pformat(values)

    @staticmethod
    def _padding_mod(st: str, offset=7, mod=4):
        """
        123 \\
        1   \\
        12312341    \\
        1231
        Args:
            strs:
            mod:

        Returns:

        """
        size = len(st)
        if size < offset:
            return st.ljust(offset, ' ')

        mnum = mod - len(st) % mod
        # if mnum == 0:
        #     mnum = mod
        return st.ljust(size + mnum, ' ')

    @staticmethod
    def safe_param_repr(values: List[tuple], level=1) -> str:
        """

        Args:
            values:
            level:

        Returns:

        """
        res = [(f"{k}={BaseParams._safe_repr(v)},", anno) for k, v, anno in values]

        # res = textwrap.fill('\n'.join(res))
        res = '\n'.join([BaseParams._padding_mod(i, offset=16, mod=4) + f'  # {anno}' for i, anno in res])

        return textwrap.indent(res, '    ')

    @classmethod
    def Space(cls, **kwargs):
        return cls().from_dict(kwargs)

    def copy(self):
        res = self.__class__()
        res._param_dict = self._param_dict.copy()
        res._repeat = self._repeat
        res._bound = copy.deepcopy(self._constrain)
        res._lock = self._lock
        return res

    def arange(self, default, left=float("-inf"), right=float("inf")) -> arange_param:
        """
        Make sure some value is into some range.

        Examples:
            params.batch_size = params.arange(20,10,100)
            print(params.batch_size) # will print '20' as default.
            params.batch_size = 300 # will raise an Exception
            params.batch_size = 50
            print(params.batch_size) # will print 50

        Args:
            k: key of the value
            default: default value
            left: left interval
            right: right interval

        Returns:
            arange_param(default, left, right)
        """
        if left < default and default < right:
            return arange_param(default, left, right)
        else:
            raise BoundCheckError(f"value {default}' should in range [{left}, {right}].")

    def choice(self, *choices) -> choice_param:
        """
        Make sure some value is into some limited values.

        Examples:
            params.dataset = params.choice('cifar10','cifar100')
            print(params.dataset) # will print 'cifar10' as default.
            params.dataset = 'mnist' # will raise an Exception
            params.dataset = 'cifar100'
            print(params.dataset) # will print 'cifar100'

        Args:
            k: key of the value
            *choices: value can be used for key

        Returns:
            choice_param(choices[0], choices)


        """
        return choice_param(choices[0], choices)

    def default(self, value: Any = None, warn=True) -> default_param:
        """
        Set a default value to a key. This default value will be set only when `key` doesn't exists in Params.

        Examples:
        ```
        params.margin = 3
        params.margin = params.default(5,True)
        params.non_exists = params.default(0.3,True)
        print(params.margin)
        print(params.non_exists)
        ```

        Args:
            value: default value
            warn: warn if default value when set this default value, to warn user set this value manully in advance.
                default is True.

        Returns:
            default_param(value, warn)
        """
        return default_param(value, warn)

    def grid_search(self, key, iterable: Iterable):
        """
        Return a iterator where each element is the original Params instance
        but changes the value of variable `key` to the value in `iterable` one by one.

        Args:
            key:
            iterable:

        Returns:
            Iterator of Params instance with the same class and different value in `key`
        """
        for v in iterable:
            res = self.copy()
            res[key] = v
            yield res

    def grid_range(self, count):
        """
        Repeat this Params instance `count` times, and return its iterator.
        You can identify these instances by `params._repeat`(start from zero)

        Args:
            count:

        Returns:
            Iterator of the repeated Params instance
        """
        for i in range(count):
            res = self.copy()
            res._repeat = i
            yield res

    def from_args(self):
        """
        Load key-value from command line arguments (based on facebook-Fire).


        Examples:
            python demo.py --k=12 --v=qwe

            # demo.py
            params.from_args()
            print(params.k) # will give `12` int object
            print(params.v) # will give `qwe` string object
        """

        def func(**kwargs):
            if '_help' in kwargs:
                print(self)
                exit()
                return

            if '_json' in kwargs:
                self.from_json(kwargs['_json'])
                return

            for k, v in kwargs.items():
                try:
                    self[k]
                except:
                    warnings.simplefilter('always', NewParamWarning)
                    warnings.warn(
                        "'{}' is a new param,please check your spelling.\n it's more recommended to define in advance.".format(
                            k), NewParamWarning)
                self[k] = v

        fire.Fire(func)
        self.iparams()
        return self

    def from_yaml(self, fn):
        """
        Read params from yaml file, if file path `fn` not exist or some Exceptions are raised during load, the program won't be terminal
        but error messages will be printed in stderr.

        Args:
            fn: file path of the yaml file
        """
        if os.path.exists(fn):
            try:
                import yaml
                with open(fn, encoding='utf-8') as r:
                    res = yaml.safe_load(r)
                    for k, v in res.items():
                        self[k] = v
            except ImportError as e:
                print(
                    "from_yaml() operation will be ignored, cause you havn't install yaml, use `pip install yaml` to install it")
            except Exception as e:
                print('from_yaml() operation will be ignored, cause:')
                print(e, file=sys.stderr)
        else:
            print(f'{fn} not exists, ignored load params from yaml file, please verify the path.', file=sys.stderr)
        self.iparams()
        return self

    def from_json(self, fn):
        """
        Read params from json file, if file path `fn` not exist or some Exceptions are raised during load, the program won't be terminal
        but error messages will be printed in stderr.

        Args:
            fn: file path of the yaml file
        """
        if os.path.exists(fn):
            try:
                with open(fn, encoding='utf-8') as r:
                    res = json.load(r)
                    for k, v in res.items():
                        self[k] = v
            except Exception as e:
                print('from_json() operation will be ignored, cause:')
                print(e, file=sys.stderr)
        else:
            print(f'{fn} not exists, ignored load params from json file, please verify the path', file=sys.stderr)

        self.iparams()
        return self

    def from_dict(self, dic: dict):
        """alias of update()"""
        for k, v in dic.items():
            self[k] = v

        self.iparams()
        return self

    def to_json(self, fn: str):
        """
        write params to the specified file path in json format.
        Args:
            fn: file path to write

        Notes:
            Only hashable key and values will be saved, see `lumo.base_classes.attr.jsonify()` for details.
        """
        io.dump_json(self.inner_dict().jsonify(), fn)

    def items(self):
        """like dict.items()"""
        return self._param_dict.items()

    def keys(self):
        """like dict.keys()"""
        for k in self._param_dict:
            yield k

    def update(self, dic: dict):
        """like dict.update()"""
        for k, v in dic.items():
            self._param_dict[k] = v

        return self

    def hash(self) -> str:
        """
        Return a hash string of all key and values. See attr.hash() for details.
        """
        return self._param_dict.hash()

    def lock(self):
        """
        Cause the params behavior is not exactly equal to dict, that is, Params instance will return an empty `attr` instance
        when some value not exists.

        To align this behavior, you can call `lock()`, then nonexist key will raise an AttributeError, which is equal to dict.

        Examples:
            params = Params()
            print(params.a) # will give an empty `attr` object
            params.lock()
            print(params.b) # will raise an AttributeError.

        Notes:
            Params default state is `unlock`, so ignore this method if you don't need this feature.
        """
        self._lock = True
        return self

    def unlock(self):
        """see lock for details."""
        self._lock = False
        return self

    def inner_dict(self) -> attr:
        """Return the inner attr, which is a dict-like object that saves all params."""
        return self._param_dict

    def get(self, k, default=None):
        """like dict.get()"""
        if k in self:
            return self[k]
        else:
            return default

    def replace(self, **kwargs):
        """alias of update()"""
        return self.update(kwargs)

    def contains(self, key: str):
        """like dict.contains()"""
        return key in self

    def merge(self, *params: 'BaseParams'):
        """Merge other params values"""
        for param in params:
            self._param_dict.update(param._param_dict)

    def iparams(self):
        pass


class DistributionParams(BaseParams):
    def __init__(self):
        super().__init__()
        self.backend = 'nccl'
        self.world_size = 1
        self.num_nodes = 1
        self.local_rank = -1

    @property
    def init_method(self):
        from lumo.proc.network import find_free_network_port
        port = find_free_network_port()
        return f'tcp://localhost:{port}'

    @overload
    def init_process_group(self, backend,
                           init_method=None,
                           timeout=timedelta(minutes=30),
                           world_size=1,
                           rank=-1,
                           store=None,
                           group_name=''):
        pass

    def init_process_group(self, **kwargs):
        self.init_process_group_args = (kwargs)
        return self.init_process_group_args


class OptimParams(BaseParams, OptimMixin):

    def __init__(self):
        super().__init__()


class Params(BaseParams):
    OPTIM = OptimParams()

    # DataLoaderParams = DataLoaderParams
    class SCHE:
        Cos = schedule.CosSchedule
        Linear = schedule.LinearSchedule
        Log = schedule.LogSchedule
        Exp = schedule.ExpSchedule
        Power = schedule.PowerDecaySchedule
        Const = schedule.ConstantSchedule

        PeriodCos = schedule.PeriodCosSchedule
        PeriodHalfCos = schedule.PeriodHalfCosSchedule
        PeriodLinear = schedule.PeriodLinear
        PeriodTriangle = schedule.PeriodTriangleSchedule

        List = schedule.ScheduleList

    def __init__(self):
        super().__init__()
        self.epoch = 10
        self.eidx = 0
        self.idx = 0
        self.global_step = 0
        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.stage = self.choice('init', 'train', 'test', 'val')


ParamsType = TypeVar('ParamsType', bound=Params)


class AccelerateParams(BaseParams):

    def __init__(self):
        super().__init__()
        self.device_placement: bool = True
        self.split_batches: bool = False
        self.fp16: bool = None
        self.cpu: bool = False
        self.rng_types: Optional[List[Union[str, RNGType]]] = None
        self.kwargs_handlers: Optional[List[KwargsHandler]] = None


def disable_commit():
    params = BaseParams()
    params.nocommit = True
    return params


def use_prerain(pretrain=True, pretrain_path=None):
    params = BaseParams()
    params.pretrain = pretrain
    params.pretrain_path = pretrain_path
    return params

# class optims:
#     @staticmethod
#     def sgd():
#         pass



"""

"""

from lumo.base_classes.metaclasses import Merge

def test_merge():
    class A(metaclass=Merge):
        _item = {1:2,3:4}

    class B(A):
        _item = {5:6,7:8}

    b = B()
    assert 1 in b._item and 3 in b._item and 5 in b._item and 7 in b._item
    assert 1 in B._item and 3 in B._item and 5 in B._item and 7 in B._item




"""
"""
import numpy as np
import pandas as pd
import torch
from numbers import Number
import hashlib
import os
import inspect
from typing import List, Tuple, Union, Mapping, Sequence, Set

from lumo.proc.path import cache_dir
from . import safe_io as io


def _to_string(item):
    res = item
    try:
        res = inspect.getsource(res)
        return res
    except TypeError:
        if inspect.isclass(res):
            res = res.__name__

    if isinstance(res, torch.Tensor):
        res = res.detach().cpu().numpy()

    if isinstance(res, np.ndarray):
        res = res.view(-1)
        mask = np.linspace(0, len(res), len(res) // 2, dtype=np.int)
        res = f"{res.shape}{res.dtype}{res[mask]}{res.sum()}{res.mean()}{res.std()}"

    if isinstance(res, pd.DataFrame):
        res = f"{res.columns}{res.shape}"

    if inspect.isclass(res) or inspect.isfunction(res):
        res = f"{res.__name__}"

    if not isinstance(res, (str, Number, Mapping, Sequence, Set)):
        res = f"{res.__class__.__name__}"
    else:
        res = f"{res}{res.__class__.__name__}"

    res = str(res)

    return res


def _cache_hash(*args, **kwargs) -> str:  # TODO hash any thing
    md5 = hashlib.md5()

    for arg in args:
        md5.update(_to_string(arg).encode(encoding='utf-8'))

    keys = sorted(list(kwargs.keys()))
    for k in keys:
        v = kwargs[k]
        md5.update(k.encode())
        md5.update(_to_string(v).encode(encoding='utf-8'))

    return md5.hexdigest()


def _cache_suffix(item):
    if isinstance(item, np.ndarray):
        return 'npy'
    elif isinstance(item, torch.Tensor):
        return 'pth'
    elif isinstance(item, pd.DataFrame):
        return 'ft'
    else:
        return 'pkl'


def _drop_pandas(item: pd.DataFrame):
    for key in {'index', 'level_0'}:
        if key in item.columns:
            item = item.drop(key, 1)
    return item


def _dump_cache(item, fn):
    if isinstance(item, pd.DataFrame):
        item = _drop_pandas(item)
        item = item.reset_index()
        item = _drop_pandas(item)
        try:
            item.to_feather(fn)
        except:
            item.to_pickle(fn)
    elif isinstance(item, np.ndarray):
        np.save(fn, item)
    elif isinstance(item, torch.Tensor):
        torch.save(item, fn)
    else:
        io.dump_state_dict(item, fn)


def _load_cache(fn: str):
    if fn.endswith('ft'):
        try:
            return pd.read_feather(fn)
        except:
            return pd.read_pickle(fn)
    elif fn.endswith('npy'):
        return np.load(fn)
    elif fn.endswith('pth'):
        return torch.load(fn, map_location='cpu')
    else:
        return io.load_state_dict(fn)


def save_cache(items: Union[List, Tuple], *args, **kwargs):
    hash = _cache_hash(*args, **kwargs)
    path = os.path.join(cache_dir(), hash)
    os.makedirs(path, exist_ok=True)
    for i, item in enumerate(items):
        cfn = f"cache_{i:03d}.{_cache_suffix(item)}"
        _dump_cache(item, os.path.join(path, cfn))
    return path


def load_from_cache_path(path):
    fs = [os.path.join(path, f) for f in sorted(os.listdir(path)) if f.startswith('cache')]
    items = [_load_cache(f) for f in fs]
    return items


def load_if_exists(*args, **kwargs):
    hash = _cache_hash(*args, **kwargs)
    path = os.path.join(cache_dir(), hash)
    if os.path.isdir(path):
        items = load_from_cache_path(path)
        return items, path
    return None, None



"""

"""


from lumo import globs
from torchvision import datasets
from lumo.contrib.data import SemiDataset

base = datasets.CIFAR10(globs["nddatasets"], train=True, download=False)



SemiDataset.create_indice(4000,5000,)


import importlib
from typing import overload

from torch.optim.optimizer import Optimizer

from .attr import attr


class OptimBuilder(attr):
    # @overload
    # def __init__(self,values:):

    # def __init__(self, name: str = None, **kwargs):
    #     super().__init__()
    #     self.args = attr.from_dict(kwargs)  # type:attr
    #     self.name = name
    @property
    def args(self):
        return self

    def build(self, parameters, optim_cls=None) -> Optimizer:
        assert 'name' in self
        lname = self['name'].lower()
        args = dict(self)
        args.pop('name')
        if optim_cls is None:
            optim_lib = importlib.import_module("torch.optim.{}".format(lname))
            optim_cls = getattr(optim_lib, self.name, None)

        return optim_cls(parameters, **args)


class ScheduleParams(attr):
    def __init__(self, name: str, **kwargs):
        super().__init__()
        self.name = name
        self.args = attr.from_dict(kwargs)

    def build(self):
        pass


class ParamsFactory:
    @staticmethod
    def _filter_none(**kwargs):
        return {k: v for k, v in kwargs.items() if v is not None}

    @staticmethod
    def create_optim(name, **kwargs):
        kwargs = ParamsFactory._filter_none(**kwargs)
        return OptimBuilder(name=name, **kwargs)

    @staticmethod
    def create_scheule(name, **kwargs):
        kwargs = ParamsFactory._filter_none(**kwargs)
        return

    # opf = OptimFactory.create("SGD", lr=0.1)


class OptimMixin():
    @overload
    def create_optim(self, name='SGD', lr=None, momentum=0, dampening=0, weight_decay=0,
                     nesterov=False) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='Adam', lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0,
                     amsgrad=False) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='Adadelta', lr=1.0, rho=0.9, eps=1e-6, weight_decay=0) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='Adagrad', lr=1e-2, lr_decay=0, weight_decay=0, initial_accumulator_value=0,
                     eps=1e-10) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='AdamW', lr=2e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='AdamW',
                     lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=1e-2, amsgrad=False) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='ASGD',
                     lr=1e-2, lambd=1e-4, alpha=0.75, t0=1e6, weight_decay=0) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='LBFGS',
                     lr=1, max_iter=20, max_eval=None, tolerance_grad=1e-7, tolerance_change=1e-9, history_size=100,
                     line_search_fn=None) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='RMSprop', lr=1e-2, alpha=0.99, eps=1e-8, weight_decay=0, momentum=0,
                     centered=False) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='Rprop', lr=1e-2, etas=(0.5, 1.2), step_sizes=(1e-6, 50)) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='SparseAdam', lr=1e-3, betas=(0.9, 0.999), eps=1e-8) -> OptimBuilder:
        pass

    def create_optim(self, name=None, **kwargs) -> OptimBuilder:
        return ParamsFactory.create_optim(name=name, **kwargs)



from typing import Mapping, TypeVar, Type

T = TypeVar('T')


class K:
    @staticmethod
    def join(*args):
        return '_'.join([i for i in args if i is not None])

    @staticmethod
    def get_str(mem: Mapping, key, default=None) -> str:
        return mem.get(key, default)

    @staticmethod
    def get_type(mem: Mapping, key, default=None, *, type: Type[T]) -> T:
        return mem.get(key, default)



import os
import pickle
import time

from ..utils import random


class RndManager:
    """
    用于管理随机种子
    """

    def __init__(self, save_dir="./rnd"):
        self.save_dir = save_dir

    def mark(self, name):
        """
        用于数据集读取一类的，需要特定步骤每一次试验完全相同
        Args:
            name: 该次标记固定种子的名字，第一次调用该方法会在特定目录存放当前状态，
            第二次调用会在该位置读取当前随机种子状态

        Returns:

        """
        stt = self._get_rnd_state(name)
        if stt is not None:
            random.set_state(stt)
            return True
        else:
            self._save_rnd_state(name)
            return False

    def int_time(self):
        """用于获取一个理论上不会重复随机种子"""
        return int(str(time.time()).split(".")[-1])

    def shuffle(self, name='shuffle', seed=None):
        """
        打乱，一般用于复现试验的时候随机一个种子
        Args:
            name:
            seed:

        Returns:

        """
        if seed is None:
            random.fix_seed(self.int_time())
        else:
            random.fix_seed(seed)

    def list(self):
        """列出当前保存的所有种子"""
        return [os.path.join(self.save_dir, f) for f in os.listdir(self.save_dir) if f.endswith('rnd')]

    def _save_rnd_state(self, name):
        """保存种子"""
        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)

        seed = random.hashseed(name)
        stt = random.fix_seed(seed)

        with open(self._build_state_name(name), "wb") as f:
            pickle.dump(stt, f)

    def _have_rnd_state(self, name) -> bool:
        """判断是否存在某个种子"""
        if not os.path.exists(self.save_dir):
            return False
        return os.path.exists(self._build_state_name(name))

    def _get_rnd_state(self, name):
        """获取某个种子"""
        if not self._have_rnd_state(name):
            return None
        with open(self._build_state_name(name), "rb") as f:
            return pickle.load(f)

    def _build_state_name(self, name, replacement=False):
        if replacement:
            i = 1
            fn = os.path.join(self.save_dir, "{}.{:02d}.rnd".format(name, i))
            while os.path.exists(fn):
                i += 1
                fn = os.path.join(self.save_dir, "{}.{:02d}.rnd".format(name, i))
        else:
            fn = os.path.join(self.save_dir, "{}.rnd".format(name))

        return fn



"""
Safe means Exceptions won't be raised. When
 - call `dump_xxx` methods, True/False will be returned to indicate success/failed.
 - call `load_xxx` methods, None will be returned if something wrong happened.
"""
import json
import os
import textwrap
from collections import namedtuple
from importlib.machinery import SourceFileLoader

import torch

from lumo.decorators.safe import safe_dump, safe_load

pyval = namedtuple('pyval', ['name', 'value', 'annotation', 'commented'],
                   defaults=[None, None, False])


@safe_dump()
def dump_pyconfig(obj: list, fn: str):
    assert fn.endswith('.py')
    with open(fn, 'w', encoding='utf-8') as w:
        for (k, v, doc, commented) in obj:  # type:str
            if commented:
                commented = '# '
            else:
                commented = ''
            res = []
            if doc is not None:
                doc = str(doc)
                doc_ln = len(doc.split('\n'))
                if doc_ln > 1:
                    res.append(textwrap.indent(doc, '# '))
                    res.append('\n')

                if doc_ln == 0 or doc_ln > 1:
                    res.append(f'{commented}{k} = {v.__repr__()}\n')
                elif doc_ln == 1:
                    res.append(f'{commented}{k} = {v.__repr__()}  # {doc}\n')
            else:
                res.append(f'{commented}{k} = {v.__repr__()}\n')
            res.append('\n')
            w.write(''.join(res))
    return True


@safe_load()
def load_pyconfig(fn):
    module_name = os.path.splitext(os.path.basename(fn))[0]
    foo = SourceFileLoader(module_name, fn).load_module()
    res = {}
    for val_name in dir(foo):
        if val_name.startswith('__'):
            continue
        if isinstance(foo.__dict__[val_name], (list, dict, set, int, str, float)):
            res[val_name] = foo.__dict__[val_name]
    return res


@safe_dump()
def dump_json(obj, fn):
    with open(fn, 'w', encoding='utf-8') as w:
        json.dump(obj, w, indent=2)


@safe_dump()
def dump_yaml(obj, fn):
    import yaml
    with open(fn, 'w', encoding='utf-8') as w:
        yaml.safe_dump(obj, w)
    return fn


@safe_dump()
def dump_state_dict(obj, fn):
    torch.save(obj, fn)
    return fn


@safe_load(default={})
def load_json(fn):
    with open(fn, 'r', encoding='utf-8') as r:
        return json.load(r)


@safe_load(default={})
def load_yaml(fn):
    import yaml
    with open(fn, 'r', encoding='utf-8') as r:
        return yaml.safe_load(r)


@safe_load()
def load_state_dict(fn: str, map_location='cpu'):
    ckpt = torch.load(fn, map_location=map_location)
    return ckpt


def remove_file(fn):
    if os.path.exists(fn) and os.path.isfile(fn):
        os.remove(fn)


@safe_load()
def load_string(fn):
    with open(fn, 'r', encoding='utf-8') as r:
        return ''.join(r.readlines())


@safe_dump()
def dump_string(string: str, fn, append=False):
    mode = 'w'
    if append:
        mode = 'a'
    with open(fn, mode, encoding='utf-8') as w:
        w.write(string)
    return fn


def safe_getattr(self, key, default=None):
    try:
        return getattr(self, key, default)
    except:
        return default



from lumo.contrib import pickle
import os


def test_dump():
    fn = 'a.aa'
    try:
        with pickle.cached(fn) as cfn:
            pickle.dump(fn, cfn)
            assert os.path.exists(cfn)
            assert not os.path.exists(fn)

        assert not os.path.exists(cfn)
        assert os.path.exists(fn)

        assert (pickle.load(fn) == fn)

        assert not os.path.exists(fn)
    except BaseException as e:
        raise
    finally:
        if os.path.exists(fn):
            os.remove(fn)



"""

"""
import warnings


class BoundCheckError(BaseException):    pass


class NewParamWarning(Warning):    pass


class NoneWarning(Warning):    pass


class AttrTypeNotFoundWarning(Warning):    pass


class MeterNameException(Exception): pass


warnings.simplefilter('always', NoneWarning)
warnings.simplefilter('always', AttrTypeNotFoundWarning)



from itertools import zip_longest, chain, accumulate, repeat
from operator import add
from typing import Sized, Iterable
from typing import Sequence


def chunk(seq: Sequence, chunksize=10, pad_last=False, pad_item=None):
    ress = [iter(seq)] * chunksize

    for res in zip_longest(*ress, fillvalue=pad_item):
        if not pad_last:
            res = [i for i in res if i is not None]
        yield res


def window(seq: Sized, n: int, strid: int = 1, drop_last: bool = False):
    for i in range(0, len(seq), strid):
        res = seq[i:i + n]
        if drop_last and len(res) < n:
            break
        yield res


def window2(seq: Iterable, n: int, strid: int = 1, drop_last: bool = False):
    it = iter(seq)
    result = []
    step = 0
    for i, ele in enumerate(it):
        result.append(ele)
        result = result[-n:]
        if len(result) == n:
            if step % strid == 0:
                yield result
            step += 1
    if not drop_last:
        yield result


def lfilter(func, iterable):
    return list(filter(func, iterable))


def lmap(func, iterable):
    return list(map(func, iterable))


def lchain(*iterable):
    return list(chain(*iterable))


class safe_cycle():
    class _cycle_none():
        pass

    def __init__(self, iter_item):
        self.item = iter_item
        self.iter = None

    def __next__(self):
        if self.iter is None:
            self.iter = iter(self.item)
            cur = next(self.iter, safe_cycle._cycle_none())
            if isinstance(cur, safe_cycle._cycle_none):
                raise StopIteration('item has no element.')
            return cur
        else:
            cur = next(self.iter, safe_cycle._cycle_none())
            if not isinstance(cur, safe_cycle._cycle_none):
                return cur
            else:
                self.iter = iter(self.item)
                return next(self)

    def __iter__(self):
        while True:
            yield next(self)


def accumulate_slice(iterable):
    """
    ```
    for sl in list(accumulate_slice(range(10))):
        print(list(range(45)[sl]))

    []
    [0]
    [1, 2]
    [3, 4, 5]
    [6, 7, 8, 9]
    [10, 11, 12, 13, 14]
    [15, 16, 17, 18, 19, 20]
    [21, 22, 23, 24, 25, 26, 27]
    [28, 29, 30, 31, 32, 33, 34, 35]
    [36, 37, 38, 39, 40, 41, 42, 43, 44]

    ```
    Args:
        iterable:

    Returns:

    """
    offset, acc = repeat(iterable, 2)
    for a, b in zip(offset, accumulate(acc, add)):
        yield slice(b - a, b)



"""
This module aims to provide an easy and general method to generate series values, which can be used in
 applying learning rate of optimizer or scaling a loss value as weight.

To meet this demand, the concept 'schedule' is summarized as a periodic math function, which have
left/right interval and  start/end value.

Hence, a `Schedule` class is provided, it receives four base parameters: start, end, left, and right corresponding to
the above concept respectively.

This class provides some common methods. For example, when you have a schedule instance, you can apply learning rate by
simple call `schedule.scale()` or `schedule.apply()` function.

And, you can use `schedule.plot()` to plot a curve of the values in each step. The plot function use `matplotlib` as backend.

If you don't want to create an instance, you can call classmethod `get(cls, cur, start=0, end=1, left=0, right=1)` to get
value.

Except for the base `Schedule` class, some other subclasses of `Schedule`  which may be general used is provided, too. All can
 be easily understand by their names and plot curves.
"""
import numpy as np
from lumo.base_classes.attr import attr
from functools import lru_cache
from typing import List


class Schedule(attr):
    """
    ratio 变化为从 1 - 0
    """

    # __schdule_name__ = None

    def toggle_constant(self, toggle=True):
        """fix the schedule as the first value"""
        self.constant = toggle
        return self

    def __repr__(self):
        content = ', '.join(["{}={}".format(k, v) for k, v in self.items()])
        return "{}({})".format(self.__class__.__name__, content)

    def __call__(self, cur):
        raise NotImplementedError()

    def plot(self, num=1000, left=0, right=1000, show=True):
        """
        Plot a curve of the schedule.
        Args:
            num:
            left:
            right:

        Returns:
            plt.plot

        Notes:
            You may need to call `plt.show`() to show it.
        """
        from matplotlib import pyplot as plt

        x = np.linspace(left, right, num)
        y = [self(i) for i in x]

        res = plt.plot(x, y)
        plt.title(str(self))
        if show:
            plt.show()
        return res

    def scale(self, optimizer, cur):
        """
        Scale the learning rate by current value. 'scale' means not apply the current schedule value directly to
        the learning rate, but multiple the initial learning rate. You can use `schedule.apply()` to apply the schedule
        value directly.

        Notes:
        -------
        When first apply scale function, a `_raw_lr` represent initial lr will be set in each param_group, then, the
        learning rate(store in param_groups with key 'lr') will be calculated by `_raw_lr * schedule(cur)`.

        Args:
            optimizer: A pytorch optimizer instance.
            cur: current step of this schedule.

        Returns:
            Current schedule value.
        """
        ratio = self(cur)
        for param_group in optimizer.param_groups:  # type:dict
            raw_lr = param_group.setdefault('_raw_lr', param_group['lr'])
            param_group['lr'] = raw_lr * ratio

        return ratio

    def apply(self, optimizer, cur):
        """
        Apply the learning rate with current schedule value.

        Args:
            optimizer: A pytorch optimizer instance.
            cur: current step of this schedule.

        Returns:

        """
        new_lr = self(cur)
        for param_group in optimizer.param_groups:  # type:dict
            param_group['lr'] = new_lr

        return new_lr


class ContinuousSche(Schedule):
    def __init__(self, start=0., end=1., left=0, right=1, *args, **kwargs):
        super().__init__()
        self.left = left
        self.right = right
        self.start = start
        self.end = end
        self.constant = False
        assert left != right

    def ratio(self, cur):
        if self.constant:
            return 0
        return (cur - self.left) / (self.right - self.left)

    @classmethod
    def get_val(cls, cur, start=0, end=1, left=0, right=1, *args, **kwargs):
        """get the current schedule value without create `schedule` instance. """
        return cls(start=0, end=1, left=0, right=1, *args, **kwargs)(cur)

    def plot(self, num=1000, left=None, right=None, show=True):
        if left is None:
            left = self.left

        if right is None:
            right = self.right

        return super().plot(num, left, right, show)


class PeriodSchedule(Schedule):
    """
    period
    """

    def __init__(self, start=0, end=1, left=0, period=1, *args, **kwargs):
        super().__init__()
        self.left = left
        self.period = period
        self.start = start
        self.end = end
        self.constant = False

    def ratio(self, cur):
        if self.constant:
            return 0
        if cur < self.left:
            in_period = (self.period - (self.left - cur) % self.period) % self.period
        else:
            in_period = float(cur - self.left) % self.period
        return in_period / self.period

    @classmethod
    def get_val(cls, cur, start=0, end=1, left=0, right=1, *args, **kwargs):
        """get the current schedule value without create `schedule` instance. """
        return cls(start=start, end=end,
                   left=left, right=right, *args, **kwargs)(cur)

    def plot(self, num=1000, left=None, n_period=5, show=True):
        if left is None:
            left = self.left

        right = self.period * n_period

        return super().plot(num, left, right, show)


class CosSchedule(ContinuousSche):
    """one cycle cosine functoin"""

    def __call__(self, cur):
        if self.constant:
            return self.start

        if cur < self.left:
            return self.start
        elif cur > self.right:
            return self.end

        ratio = self.ratio(cur)
        cos_ratio = 0.5 * (1 + np.cos(ratio * np.pi))
        return self.start * cos_ratio + self.end * (1 - cos_ratio)


class ConstantSchedule(ContinuousSche):
    def __init__(self, value=0.5, *args, **kwargs):
        super().__init__(start=value, end=value, *args, **kwargs)

        self.constant = True


class LinearSchedule(ContinuousSche):
    """linear schedule"""

    def __call__(self, cur):
        if self.constant:
            return self.start

        if cur < self.left:
            return self.start
        elif cur > self.right:
            return self.end

        linear_ratio = self.ratio(cur)
        return self.start * (1 - linear_ratio) + self.end * linear_ratio


class ExpSchedule(ContinuousSche):
    """slow to quick"""

    def __call__(self, cur):
        if self.constant:
            return self.start

        if cur < self.left:
            return self.start
        elif cur > self.right:
            return self.end

        ratio = self.ratio(cur)
        residual = np.exp(-5)

        exp_ratio = np.exp((ratio - 1) * 5) - residual * (1 - ratio)
        return self.start * (1 - exp_ratio) + self.end * exp_ratio


class LogSchedule(ContinuousSche):
    """quick to slow"""

    def __call__(self, cur):
        if self.constant:
            return self.start

        if cur < self.left:
            return self.start
        elif cur > self.right:
            return self.end

        ratio = self.ratio(cur)

        residual = np.exp(-5)

        log_ratio = 1 - np.exp(-ratio * 5) + residual * ratio
        return self.start * (1 - log_ratio) + self.end * log_ratio


class PeriodCosSchedule(PeriodSchedule):
    """
    periodic cosine schedule
    """

    def __call__(self, cur):
        ratio = self.ratio(cur)
        cos_ratio = 0.5 * (1 + np.cos(ratio * np.pi * 2))
        return self.start * cos_ratio + self.end * (1 - cos_ratio)


class PeriodHalfCosSchedule(PeriodSchedule):
    """
    half periodic cosine schedule, period is (right-left)
    """

    def __call__(self, cur):
        # cur = float(cur - self.left) % (self.right - self.left)
        ratio = self.ratio(cur)
        cos_ratio = 0.5 * (1 + np.cos(ratio * np.pi))
        return self.start * cos_ratio + self.end * (1 - cos_ratio)


class PeriodTriangleSchedule(PeriodSchedule):
    def __init__(self, start=0, end=1, left=0, left_period=1, right_period=1, *args, **kwargs):
        super().__init__(start=0, end=1, left=0,
                         period=(left_period + right_period),
                         *args, **kwargs)
        assert left_period > 0 and right_period > 0
        self.left_period = left_period
        self.right_period = right_period

    def __call__(self, cur):
        ratio = self.ratio(cur)
        mid_ratio = self.left_period / (self.right_period + self.left_period)
        # if ratio < mid_ratio:
        #     return self.
        if ratio < mid_ratio:
            ratio = ratio / mid_ratio
            return self.start * (1 - ratio) + self.end * ratio
        else:
            ratio = (ratio - mid_ratio) / (1 - mid_ratio)
            return self.end * (1 - ratio) + self.start * ratio


class PeriodLinear(PeriodSchedule):
    """
    sawtooth wave, like a period line schedule
    """

    def __call__(self, cur):
        ratio = self.ratio(cur)
        return self.start * (1 - ratio) + self.end * ratio


class PowerDecaySchedule(Schedule):
    """equal to tf.train.exponential_decay, decay every <decay_steps> with a base of <decay_rate> """

    def __init__(self, start, decay_steps, decay_rate, end=None):
        super().__init__()
        self.start = start
        self.decay_steps = decay_steps
        self.decay_rate = decay_rate
        self.end = end

    def __call__(self, cur):
        rate = np.power(self.decay_rate, cur // self.decay_steps)
        res = self.start * rate
        if self.end is not None:
            res = max(self.end, res)

        return res


class ScheduleList(Schedule):
    def __init__(self, schedules: List[Schedule], bound='left'):
        super().__init__()
        if len(schedules) > 0:
            self.bound = bound
            if bound == 'left':
                self.schedules = sorted(schedules, key=lambda x: x.left)
            elif bound == 'right':
                self.schedules = sorted(schedules, key=lambda x: x.right)
            else:
                assert False

            self.left = self.schedules[0].left
            self.right = self.schedules[-1].right
        # super().__init__(None, None, left, right, *args, **kwargs)

    def __reduce__(self):
        return (self.__class__, (self.schedules, self.bound))

    def __call__(self, cur):
        if self.bound == 'left':
            for i, schedule in enumerate(self.schedules):
                if i + 1 < len(self.schedules):
                    if cur < self.schedules[i + 1].left:
                        return schedule(cur)
                    else:
                        continue
                else:
                    return schedule(cur)
        elif self.bound == 'right':
            schedule = self.schedules[-1]

            for i, schedule in enumerate(self.schedules):
                if cur < self.schedules[i].right:
                    return schedule(cur)
            return schedule(cur)
        else:
            assert False

    def __repr__(self):
        content = ', '.join([i.__class__.__name__ for i in self.schedules])
        return '{}({})'.format(self.__class__.__name__, content)

    def plot(self, num=1000, left=None, right=None, show=True):
        if left is None:
            left = self.left

        if right is None:
            right = self.right

        return super().plot(num, left, right, show)



import os
from torch import distributed as dist


def local_rank():
    rank = int(os.environ.get('LOCAL_RANK', '-1'))
    if rank == -1:
        if dist.is_available() and dist.is_initialized():
            rank = dist.get_rank()

    return rank


def world_size():
    size = int(os.environ.get('WORLD_SIZE', '0'))
    if size == 0:
        if dist.is_available() and dist.is_initialized():
            size = dist.get_world_size()
    return size


def is_dist():
    return local_rank() >= 0



import functools
import time
import torch
import os
import subprocess
from . import re
from lumo.base_classes.tree import tree
from functools import partial

match_mem = re.compile('([0-9]+) +([0-9]+)[^|]* ([0-9]+)MiB')


class DeviceMem():
    def __init__(self):
        self.line_mem = tree()

    def _parse_device_pid_mem_pair(self, lines):
        for lid, line in enumerate(lines):
            res = re.search(match_mem, line)
            if res is not None:
                _device, _pid, _mib = [int(i) for i in res.groups()]
                self.line_mem[_device][_pid] = lid
                yield _device, _pid, _mib

    def try_parse(self, lines, pid, device):
        """ try parse mem from cached lid directly.
        Returns:
             -1 means failed.
             others means successd and its memory.

        """
        lid = self.line_mem[device][pid]
        if isinstance(lid, dict):
            return -1
        elif lid > len(lines):
            return -1
        else:
            res = re.search(match_mem, lines[lid])
            if res is None:
                return -1
            else:
                _device, _pid, _mib = [int(i) for i in res.groups()]
                if _pid == pid and _device == device:
                    return _mib
                else:
                    return -1

    def re_parse(self, lines, pid, device):
        _res = self.try_parse(lines, pid, device)
        if _res != -1:
            return _res

        for _device, _pid, _mib in self._parse_device_pid_mem_pair(lines):
            if _pid == pid and _device == device:
                return _mib

        return 0

    def _get_nvidia_smi(self):
        proc = subprocess.Popen(['nvidia-smi'], stdout=subprocess.PIPE)
        lines = proc.stdout.readlines()
        return [i.decode() for i in lines]

    def _device_equal(self, da, db):
        if isinstance(da, (int, str)):
            da = torch.device(da)
        if isinstance(db, (int, str)):
            db = torch.device(db)
        return da == db

    def get_device_release_mem(self, device):
        """ get device memory left."""
        s_pid = os.getpid()
        total = self.get_device_mem(device)
        for _device, _pid, _mib in self._parse_device_pid_mem_pair(self._get_nvidia_smi()):
            if self._device_equal(device, _device):
                total -= _mib

        return total

    def get_device_mem(self, device):
        """  returns device total memory(unit: MB)        """
        return torch.cuda.get_device_properties(device).total_memory // (1024 * 1024)

    def get_pid_device_mem(self, pid, device):
        """
        尽可能有效率的得到进程在某设备下占用的显存（通过命令行程序调用获取）
        :param pid:
        :param device:
        :return:
        """
        if isinstance(device, torch.device):
            device = device.index
        elif isinstance(device, str):
            device = torch.device(device).index

        lines = self._get_nvidia_smi()
        mib = self.try_parse(lines, pid, device)
        if mib == -1:
            mib = self.re_parse(lines, pid, device)

        return mib


_memer = DeviceMem()
_pid = os.getpid()

if torch.cuda.is_available():
    get_pid_device_mem = partial(_memer.get_pid_device_mem, pid=_pid, device=torch.cuda.current_device())


class memory(object):
    r"""
    优雅的抢卡
    Args:
        memory: 需要占用的内存，以 MB 为单位
        device: 需要占用内存的设备
        hold:
        unit:
    Example::
        >>> import lumo
        >>> with lumo.memory(5000):
        ...   y = x * 2

        >>> @lumo.memory(1024)
        ... def doubler(x):
        ...     ...

        >>> lumo.memory(10000).start()
        ... # do something

    Why use nvidia-smi to get memory useage? see:
        https://github.com/pytorch/pytorch/issues/12873
    """

    def __init__(self, memory, device=None, hold=False) -> None:
        super().__init__()
        if device is None:
            device = torch.cuda.current_device()
        if isinstance(device, (str, int)):
            device = torch.device(device)

        self.need = memory
        self.device = device
        self.hold = hold
        self.exc_time = 0
        self.acc = 5
        self.mem = []
        self.last_success = _memer.get_pid_device_mem(_pid, self.device)

    def copy(self, pid: int, wait: bool = True):
        self.need = _memer.get_pid_device_mem(pid, self.device)
        if wait:
            self.wait(pid)
        else:
            self.start()

    def wait(self, pid):
        while _memer.get_pid_device_mem(pid, self.device) > 0:
            time.sleep(0.5)
        self.start()

    def immediately(self, pre_init=False):
        """
        等待，直到内存有空间后，开始申请相应显存，优雅，礼貌，推荐
        Args:
            pre_init: 是否初始化 CUDA（这将在一开始消耗一定显存），默认为 False，即不抢占任何内存，
                直到设备释放足够空间后开始抢占。
        """
        while True:
            _left = _memer.get_device_release_mem(self.device)
            _allocated = _memer.get_pid_device_mem(_pid, self.device)

            if pre_init and _allocated == 0:
                self._malloc(1, init=True)
                continue

            _need = self.need - _allocated
            if _need < 0:
                return self.end()

            if _left > _need:
                if _allocated == 0:
                    self._malloc(1, init=True)
                    continue

                res = self._malloc(_need)
                time.sleep(0.5)
                if res:
                    return self.end()

            print("need {}Mb, {}Mb allocated, "
                  "waiting for {}Mb released, "
                  "but only {}Mb left.".format(self.need,
                                               _allocated, _need,
                                               _left), end='\r')

    def _malloc(self, size, init=False):
        """ unit: mb """
        try:
            tmp = torch.rand(size, 1048576 // 4, device=self.device)
            if not init:
                self.mem.append(tmp)
            return True
        except Exception as e:
            return False

    def end(self):
        print()
        if self.hold:
            print('press keyboardinterrupt to end')
            try:
                while True:
                    # do some Fake thing
                    self.mem[-1].random_()
                    time.sleep(0.1)
            except KeyboardInterrupt:
                print('continue')

    def start(self, immediately=True):
        if immediately:
            self.immediately()
        else:
            self.invade()

    def invade(self, unit=5):
        """一点一点的侵占，有多少占用多少，直到申请满为止，比较粗鲁，不友好，不推荐"""
        try:
            while self.last_success < self.need:
                res = self._malloc(unit + self.acc)
                if res:
                    self.acc += unit
                    self.last_success = _memer.get_pid_device_mem(_pid, self.device)
                    time.sleep(0.1)
                else:
                    self.exc_time += 1
                    self.acc = max(self.acc - unit, 0)
                    time.sleep(0.5)
                print('{}/{}Mb, try={}, pid={}'.format(self.last_success,
                                                       self.need,
                                                       self.exc_time,
                                                       os.getpid()), end='\r')
            self.end()
        except KeyboardInterrupt:
            print('\nabort.')

    def __enter__(self):
        self.invade()

    def __exit__(self, *args):
        del self.mem[:]
        torch.cuda.empty_cache()
        return True

    def __call__(self, func):
        @functools.wraps(func)
        def decorate_no_grad(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorate_no_grad

    @staticmethod
    def hold_current():
        count = torch.cuda.device_count()
        mems = [_memer.get_pid_device_mem(_pid, i) for i in range(count)]
        for i, mem in enumerate(mems):
            if mem > 0:
                memory(mem, device=i, hold=(i == count - 1)).start()


if __name__ == '__main__':
    memory(11000, 3).immediately(False)



"""
 - 图 Dataset， gg['item_id'] -> gg['words'] -> gg['item_id'] ，先随机采样，然后尝试直接聚类的方式？
 - Bart 模型 + p-tuning
"""
from lumo import Trainer, Logger, Params, DatasetBuilder, DataModule, K, callbacks
from lumo.contrib.data.collate import CollateBase
from lumo.base_classes import TrainerStage
from lumo.kit import ParamsType


class MultiPM(Params):

    def __init__(self):
        super().__init__()
        self.epoch = 20
        self.batch_size = 128
        self.num_workers = 8
        self.pin_memory = True

        self.pretrain = True
        self.pretrain_path = None

    def iparams(self):
        super().iparams()
        self.lr_sche = self.SCHE.Cos(start=0, end=self.epoch,
                                     left=self.optim.args.lr,
                                     right=1e-6)


class Key(K):
    pass


K = Key


# DataModule
class GlobalTransform():
    def __init__(self, params: MultiPM):
        self.params = params

    def __call__(self, mem: dict):
        return mem


def mtds(params: MultiPM):
    """
    视频路径

    :param params:
    :return:
    """

    ds = (
        DatasetBuilder()
            # .add_input()
            # .add_output()
            .add_global_transform(GlobalTransform(params))
    )

    return ds


class MTCollate(CollateBase):

    def initial(self, params: MultiPM):
        pass


class MTDM(DataModule):

    def idataloader(self, params: MultiPM, stage: TrainerStage, repeat: bool = False):
        ds = mtds(params)

        if stage.is_train:
            ds.subset(range(len(ds) * 4 // 5))
        elif stage.is_eval:
            ds.subset(range(len(ds) * 4 // 5, len(ds)))

        loader = ds.DataLoader(batch_size=params.batch_size,
                               num_workers=params.num_workers,
                               pin_memory=params.pin_memory,
                               collate_fn=CollateBase())

        self.regist_dataloader_with_stage(stage, loader)


# Trainer


class MTTrainer(callbacks.TrainCallback,
                Trainer):
    def icallbacks(self, params: ParamsType):
        callbacks.LoggerCallback(step_frequence=1).hook(self)
        callbacks.LRSchedule().hook(self)  # auto get params.lr_sche to apply lr rate
        callbacks.AutoLoadModel().hook(self)

        if isinstance(self, callbacks.TrainCallback):
            self.hook(self)

    def imodels(self, params: ParamsType):
        super().imodels(params)

    def train_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        pass

    def evaluate_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        pass


def main():
    pm = MultiPM()
    pm.from_args()

    dm = MTDM()

    tr = MTTrainer(pm)
    tr.train(dm)


if __name__ == '__main__':
    main()



"""

"""

from lumo.frame.experiment import globs

globs["nddatasets"] = "E:/Download/dataset"

globs.time = 0

print()
# globs.update("local", "a", 1)
# globs.list_config(True, True)

# print(globs.glob_fn)
